class DoublePageDefenitionError(Exception):
    """
    Исключение возникает, если обе одноименных страницы
    отнаследованы от одного и того же класса.
    """
    pass


class UrlAvailabilityError(Exception):
    """
    Исключение возникает, если адрес страницы не определен в аттрибутах
    класса и не передан с аргументами при создании объекта класса.
    """
    pass


class UrlDisparityError(Exception):
    """
    Исключение возникает, если адрес страницы указан и в аттрибутах класса, и передан
    с аргументами, но они отличаются.
    """
    pass
